import unittest


class CacheTestCase(unittest.TestCase):
    def setUp(self):
        from pathlib import Path
        from tempfile import mkdtemp
        import restage.cache
        from restage.database import Database
        from mccode_antlr.loader import parse_mcstas_instr
        self.db_dir = Path(mkdtemp())
        self.db_file = self.db_dir.joinpath('test_database.db')
        self.db = Database(self.db_file)

        self.orig_db = restage.cache.FILESYSTEM.db_write
        restage.cache.FILESYSTEM.db_write = self.db

        contents = """DEFINE INSTRUMENT simple_test_instrument(
                    par1, double par2, int par3, par4=1, string par5="string", double par6=6.6
                    )
                TRACE
                COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
                COMPONENT sample = Arm() AT (0, 0, 1) ABSOLUTE
                COMPONENT detector = Arm() AT (0, 0, 2) ABSOLUTE
                END
                """
        self.instr = parse_mcstas_instr(contents)

    def tearDown(self) -> None:
        import restage.cache
        restage.cache.FILESYSTEM.db_write = self.orig_db
        del self.orig_db

        del self.db
        if self.db_file.exists():
            self.db_file.unlink()
        del self.db_file
        if self.db_dir.exists():
            self.db_dir.rmdir()
        del self.db_dir

    def _check_attrs(self, obj, expected_values):
        for k, v in expected_values.items():
            self.assertTrue(hasattr(obj, k))
            self.assertEqual(getattr(obj, k), v)

    def test_simple_instr_file(self):
        from restage import InstrEntry
        from restage.cache import cache_instr
        import mccode_antlr
        mccode_version = mccode_antlr.__version__
        file_contents = str(self.instr)  # ideally this would be contents, but the parsed representation adds things
        binary_path = '/not/a/real/binary/path'
        retrieved = cache_instr(self.instr, mccode_version=mccode_version, binary_path=binary_path)
        self.assertTrue(isinstance(retrieved, InstrEntry))
        self._check_attrs(retrieved, {'file_contents': file_contents, 'binary_path': binary_path,
                                      'mccode_version': mccode_version})
        from_db = self.db.retrieve_instr_file(retrieved.id)
        self.assertEqual(len(from_db), 1)
        self._check_attrs(from_db[0], {'file_contents': file_contents, 'binary_path': binary_path,
                                       'mccode_version': mccode_version})

    def test_nexus_structure(self):
        pass

    def test_simulation_table(self):
        from restage import SimulationTableEntry, SimulationEntry
        from restage.instr import collect_parameter
        from restage.cache import cache_instr, cache_simulation_table
        import mccode_antlr
        mccode_version = mccode_antlr.__version__
        binary_path = '/not/a/real/binary/path'
        instr_entry = cache_instr(self.instr, mccode_version=mccode_version, binary_path=binary_path)

        sim_entry = SimulationEntry(collect_parameter(self.instr))
        sim_table = cache_simulation_table(instr_entry, sim_entry)

        self.assertTrue(isinstance(sim_table, SimulationTableEntry))
        self._check_attrs(sim_table, {'id': instr_entry.id, 'name': f'pst_{instr_entry.id}',
                                        'parameters': list(sim_entry.parameter_values.keys())})
        from_db = self.db.retrieve_simulation_table(instr_entry.id)
        self.assertEqual(len(from_db), 1)
        self._check_attrs(from_db[0], {'id': instr_entry.id, 'name': f'pst_{instr_entry.id}',
                                       'parameters': list(sim_entry.parameter_values.keys())})

    def test_simulation(self):
        from restage.tables import SimulationEntry, best_simulation_entry_match
        from restage.instr import collect_parameter
        from restage.cache import cache_instr, cache_simulation, cache_has_simulation, cache_get_simulation
        import mccode_antlr
        mccode_version = mccode_antlr.__version__
        binary_path = '/not/a/real/binary/path'
        instr_entry = cache_instr(self.instr, mccode_version=mccode_version, binary_path=binary_path)

        # double, double, int, double, string, double
        par0 = collect_parameter(self.instr, par1=1.1, par2=2.2, par3=3, par4=4.4, par5='five', par6=6.6)
        par1 = collect_parameter(self.instr, par1=1.9, par2=2.2, par3=3, par4=4.4, par5='five', par6=6.6)
        par2 = collect_parameter(self.instr, par1=1.1, par2=2.9, par3=3, par4=4.4, par5='five', par6=6.6)
        par3 = collect_parameter(self.instr, par1=1.1, par2=2.2, par3=4, par4=4.4, par5='five', par6=6.6)
        par4 = collect_parameter(self.instr, par1=1.1, par2=2.2, par3=3, par4=4.9, par5='five', par6=6.6)
        par5 = collect_parameter(self.instr, par1=1.1, par2=2.2, par3=3, par4=4.4, par5='six', par6=6.6)
        par6 = collect_parameter(self.instr, par1=1.1, par2=2.2, par3=3, par4=4.4, par5='five', par6=6.9)
        pars = (par0, par1, par2, par3, par4, par5, par6)
        for par in pars:
            cache_simulation(instr_entry, SimulationEntry(par))  # automatically inserts the table if it doesn't exist
        for par in pars:
            self.assertTrue(cache_has_simulation(instr_entry, SimulationEntry(par)))
            self.assertEqual(len(cache_get_simulation(instr_entry, SimulationEntry(par))), 1)

        # Now 'repeat' simulations with a specified seed
        for seed, par in enumerate(pars):
            cache_simulation(instr_entry, SimulationEntry(par, seed=seed))
        for seed, par in enumerate(pars):
            self.assertTrue(cache_has_simulation(instr_entry, SimulationEntry(par)))
            self.assertTrue(cache_has_simulation(instr_entry, SimulationEntry(par, seed=seed)))
            self.assertEqual(len(cache_get_simulation(instr_entry, SimulationEntry(par))), 2)
            self.assertEqual(len(cache_get_simulation(instr_entry, SimulationEntry(par, seed=seed))), 1)

        # and again where we've specified a number of particles
        for n, par in enumerate(pars):
            cache_simulation(instr_entry, SimulationEntry(par, ncount=10000+n))
        for n, par in enumerate(pars):
            self.assertTrue(cache_has_simulation(instr_entry, SimulationEntry(par)))
            matches = cache_get_simulation(instr_entry, SimulationEntry(par))
            self.assertEqual(len(matches), 3)
            best = best_simulation_entry_match(matches, SimulationEntry(par))
            self.assertEqual(best.ncount, 10000+n)


if __name__ == '__main__':
    unittest.main()



import unittest

"""
Test the multi-database mechanism by
    1. creating a temporary database and location for binaries and simulations
    2. adding at least one instrument and one simulation to the database
    3. making the first temporary database read-only, adding a second writable one
    4. 'using' the read-only simulation
    5. adding a new simulation for the instrument in the read-only database to the
       second database 
"""

class ROCacheTestCase(unittest.TestCase):
    def setUp(self):
        from pathlib import Path
        from tempfile import mkdtemp
        import restage.cache
        import mccode_antlr
        from restage.cache import cache_instr, cache_simulation_table, cache_simulation
        from restage.database import Database
        from restage.instr import collect_parameter
        from restage import SimulationEntry
        from mccode_antlr.loader import parse_mcstas_instr

        database_name = self.id().split('.')[-1] + '.db'

        self.ro_dir = Path(mkdtemp())
        self.ro_db_file = self.ro_dir.joinpath('ro_' + database_name)
        self.ro_db = Database(self.ro_db_file)

        self.orig_ro_db = restage.cache.FILESYSTEM.db_fixed
        self.orig_rw_db = restage.cache.FILESYSTEM.db_write
        restage.cache.FILESYSTEM.db_write = self.ro_db

        contents = """DEFINE INSTRUMENT simple_test_instrument(
                            par1, double par2, int par3, par4=1, string par5="string", double par6=6.6
                            )
                        TRACE
                        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
                        COMPONENT sample = Arm() AT (0, 0, 1) ABSOLUTE
                        COMPONENT detector = Arm() AT (0, 0, 2) ABSOLUTE
                        END
                        """
        self.instr = parse_mcstas_instr(contents)

        # Set up the 'read-only' part (this functionality checked in CacheTestCase)
        instr_entry = cache_instr(self.instr, mccode_version=mccode_antlr.__version__, binary_path=self.ro_dir / "bin" / "blah")
        cache_simulation_table(instr_entry, SimulationEntry(collect_parameter(self.instr)))
        self.par = collect_parameter(self.instr, par1=1., par2=2., par3=3, par4=4., par5='five', par6=6.)
        cache_simulation(instr_entry, SimulationEntry(self.par))

        # Close the database file, and re-open it read-only
        del self.ro_db
        self.ro_db = Database(self.ro_db_file, readonly=True)
        restage.cache.FILESYSTEM.db_fixed = (self.ro_db,)

        # Make a new writable database
        self.rw_dir = Path(mkdtemp())
        self.rw_db_file = self.rw_dir.joinpath('rw_' + database_name)
        self.rw_db = Database(self.rw_db_file)
        restage.cache.FILESYSTEM.db_write = self.rw_db

    def tearDown(self):
        import restage.cache
        restage.cache.FILESYSTEM.db_fixed = self.orig_ro_db
        restage.cache.FILESYSTEM.db_write = self.orig_rw_db

        del self.ro_db
        del self.rw_db
        for file in (self.ro_db_file, self.rw_db_file):
            file.unlink(missing_ok=True)
        for directory in (self.ro_dir, self.rw_dir):
            if directory.exists():
                directory.rmdir()
            del directory

    def test_ro_simulation_retrieval(self):
        from pathlib import Path
        from restage import SimulationEntry
        from restage.cache import cache_get_instr, cache_has_simulation, cache_get_simulation
        instr_entry = cache_get_instr(self.instr)
        self.assertEqual(Path(instr_entry.binary_path), self.ro_dir / "bin" / "blah")
        self.assertTrue(cache_has_simulation(instr_entry, SimulationEntry(self.par)))
        self.assertEqual(len(cache_get_simulation(instr_entry, SimulationEntry(self.par))), 1)

    def test_rw_simulation_insertion(self):
        from pathlib import Path
        from restage import SimulationEntry
        from restage.cache import FILESYSTEM
        from restage.cache import (
            cache_get_instr, cache_has_simulation, cache_get_simulation,
            cache_simulation, cache_simulation_table
        )
        from restage.instr import collect_parameter
        instr_entry = cache_get_instr(self.instr)
        self.assertEqual(Path(instr_entry.binary_path), self.ro_dir / "bin" / "blah")

        par = collect_parameter(self.instr, par1=2., par2=3., par3=4, par4=5., par5='six', par6=7.)
        entry = SimulationEntry(par)

        self.assertTrue(cache_has_simulation(instr_entry, SimulationEntry(self.par)))
        self.assertFalse(cache_has_simulation(instr_entry, entry))

        cache_simulation(instr_entry, entry)

        self.assertEqual(len(cache_get_simulation(instr_entry, entry)), 1)

        table = cache_simulation_table(instr_entry, entry)
        self.assertEqual(len(FILESYSTEM.db_write.retrieve_simulation(table.id, entry)), 1)
        self.assertEqual(len(FILESYSTEM.db_fixed[0].retrieve_simulation(table.id, entry)), 0)


import os
from unittest import TestCase
from unittest.mock import patch
from importlib import reload
import restage.config


class SettingsTests(TestCase):
    import os
    @patch.dict(os.environ, {"RESTAGE_CACHE": "/tmp/some/location"})
    def test_restage_cache_config(self):
        reload(restage.config)
        from restage.config import config
        self.assertTrue(config['cache'].exists())
        self.assertEqual(config['cache'].as_str(), '/tmp/some/location')

    @patch.dict(os.environ, {"RESTAGE_FIXED": "/tmp/some/location"})
    def test_restage_single_fixed_config(self):
        reload(restage.config)
        from restage.config import config
        self.assertTrue(config['fixed'].exists())
        self.assertEqual(config['fixed'].as_str(), '/tmp/some/location')

    @patch.dict(os.environ, {'RESTAGE_FIXED': '/tmp/a /tmp/b /tmp/c'})
    def test_restage_multi_fixed_config(self):
        reload(restage.config)
        from restage.config import config
        self.assertTrue(config['fixed'].exists())
        more = config['fixed'].as_str_seq()
        self.assertEqual(len(more), 3)
        self.assertEqual(more[0],'/tmp/a')
        self.assertEqual(more[1],'/tmp/b')
        self.assertEqual(more[2],'/tmp/c')

    @patch.dict(os.environ, {"RESTAGE_FIXED": ''})
    def test_restage_none_fixed_config(self):
        reload(restage.config)
        from restage.config import config
        from confuse.exceptions import ConfigTypeError
        self.assertTrue(config['fixed'].exists())
        self.assertTrue(config['fixed'].get() is None)
        self.assertRaises(ConfigTypeError, config['fixed'].as_str_seq)

    def test_missing_database_in_fixed_config(self):
        from tempfile import TemporaryDirectory
        from pathlib import Path
        with TemporaryDirectory() as tmpdir:
            self.assertFalse((Path(tmpdir) / 'database.db').exists())
            with patch.dict(os.environ, {'RESTAGE_FIXED': tmpdir}):
                reload(restage.config)
                from restage.config import config
                self.assertTrue(config['fixed'].exists())
                self.assertEqual(config['fixed'].as_str(), tmpdir)
                from restage.cache import FILESYSTEM
                self.assertEqual(len(FILESYSTEM.db_fixed), 0)


    def test_restage_standard_config(self):
        from os import environ
        reload(restage.config)
        from restage.config import config
        if 'cache' in config:
            self.assertEqual(config['cache'].as_str(), environ['RESTAGE_CACHE'])
        if 'fixed' in config:
            self.assertEqual(config['fixed'].as_str(), environ['RESTAGE_FIXED'])




import confuse
from os import environ
# Any platform independent configuration settings can go in 'default.yaml'
config = confuse.LazyConfig('restage', __name__)

# use environment variables specified as 'RESTAGE_XYZ' as configuration entries 'xyz'
config.set_env()
# Expected environment variables:
#   RESTAGE_FIXED="/loc/one /usr/loc/two"
#   RESTAGE_CACHE="$HOME/loc/three"


def _common_defaults():
    import yaml
    from importlib.resources import files, as_file

    common_file = files(__name__).joinpath('default.yaml')
    if not common_file.is_file():
        raise RuntimeError(f"Can not locate default.yaml in module files (looking for {common_file})")
    with as_file(common_file) as file:
        with open(file, 'r') as data:
            common_configs = yaml.safe_load(data)

    return common_configs or {}


# By using the 'add' method, we set these as the *lowest* priority. Any user/system files will override:
config.add(_common_defaults())


from __future__ import annotations


def make_scan_parser():
    from argparse import ArgumentParser
    parser = ArgumentParser('restage_scan')
    parser.add_argument('primary', nargs=1, type=str, default=None,
                        help='Primary spectrometer `.instr` file name')
    parser.add_argument('secondary', nargs=1, type=str, default=None,
                        help='Secondary spectrometer `.instr` file name')
    parser.add_argument('parameters', nargs='*', type=str, default=None)
    parser.add_argument('-R', action='append', default=[], help='Runtime parameters')
    parser.add_argument('-g', '--grid', action='store_true', default=False, help='Grid scan')
    return parser


def parse_scan():
    from .range import parse_scan_parameters
    args = make_scan_parser().parse_args()
    parameters = parse_scan_parameters(args.parameters)
    return args, parameters


def run_point(args, parameters):
    print(f'{args} {parameters}')
    pass


def entrypoint():
    """Entrypoint for the restage_scan command."""
    from .energy import bifrost_translate_energy_to_chopper_parameters
    from .range import parameters_to_scan
    args, parameters = parse_scan()
    n_points, names, scan = parameters_to_scan(parameters)
    for i, p in enumerate(scan):
        point_parameters = bifrost_translate_energy_to_chopper_parameters({k: v for k, v in zip(names, p)})
        run_point(args, point_parameters)



def _wavelength_angstrom_to_energy_mev(wavelength):
    return 81.82 / wavelength / wavelength


def get_and_remove(d: dict, k: str, default=None):
    if k in d:
        v = d[k]
        del d[k]
        return v
    return default


def one_generic_energy_to_chopper_parameters(
        calculate_choppers, chopper_names: tuple[str, ...],
        time: float, order: int, parameters: dict,
        chopper_parameter_present: bool
):
    from loguru import logger
    if any(x in parameters for x in ('ei', 'wavelength', 'lambda', 'energy', 'e')):
        if chopper_parameter_present:
            logger.warning('Specified chopper parameter(s) overridden by Ei or wavelength.')
        ei = get_and_remove(parameters, 'ei', get_and_remove(parameters, 'energy', get_and_remove(parameters, 'e')))
        if ei is None:
            wavelength = get_and_remove(parameters, 'wavelength', get_and_remove(parameters, 'lambda'))
            ei = _wavelength_angstrom_to_energy_mev(wavelength)
        choppers = calculate_choppers(order, time, ei, names=chopper_names)
        parameters.update(choppers)
    return parameters


def bifrost_translate_energy_to_chopper_parameters(parameters: dict):
    from itertools import product
    from .bifrost_choppers import calculate
    choppers = tuple(f'{a}_chopper_{b}' for a, b in product(['pulse_shaping', 'frame_overlap', 'bandwidth'], [1, 2]))
    # names = [a+b for a, b in product(('ps', 'fo', 'bw'), ('1', '2'))]
    chopper_parameter_present = False
    for name in product(choppers, ('speed', 'phase')):
        name = ''.join(name)
        if name not in parameters:
            parameters[name] = 0
        else:
            chopper_parameter_present = True
    order = get_and_remove(parameters, 'order', 14)
    time = get_and_remove(parameters, 'time', get_and_remove(parameters, 't', 170/180/(2 * 15 * 14)))
    return one_generic_energy_to_chopper_parameters(calculate, choppers, time, order, parameters, chopper_parameter_present)


def cspec_translate_energy_to_chopper_parameters(parameters: dict):
    from itertools import product
    from .cspec_choppers import calculate
    choppers = ('bw1', 'bw2', 'bw3', 's', 'p', 'm1', 'm2')
    chopper_parameter_present = False
    for name in product(choppers, ('speed', 'phase')):
        name = ''.join(name)
        if name not in parameters:
            parameters[name] = 0
        else:
            chopper_parameter_present = True
    time = get_and_remove(parameters, 'time', 0.004)
    order = get_and_remove(parameters, 'order', 16)
    return one_generic_energy_to_chopper_parameters(calculate, choppers, time, order, parameters, chopper_parameter_present)


def no_op_translate_energy_to_chopper_parameters(parameters: dict):
    return parameters


def energy_to_chopper_translator(instrument: str):
    if 'bifrost' in instrument.lower():
        return bifrost_translate_energy_to_chopper_parameters
    if 'cspec' in instrument.lower():
        return cspec_translate_energy_to_chopper_parameters
    return no_op_translate_energy_to_chopper_parameters


def get_energy_parameter_names(instr: str):
    if 'bifrost' in instr.lower():
        return ['e', 'ei', 'energy', 'wavelength', 'lambda', 'time', 't', 'order']
    elif 'cspec' in instr.lower():
        return ['e', 'ei', 'energy', 'wavelength', 'lambda', 'reps']
    else:
        return []



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from mccode_antlr.instr import Instr
from .tables import InstrEntry, SimulationTableEntry, SimulationEntry
from .database import Database

@dataclass
class FileSystem:
    root: Path
    db_fixed: tuple[Database,...]
    db_write: Database

    @classmethod
    def from_config(cls, named: str):
        from .config import config
        db_fixed = []
        db_write = None
        root = None
        if not named.endswith('.db'):
            named += '.db'
        if config['cache'].exists():
            path = config['cache'].as_path()
            if not path.exists():
                path.mkdir(parents=True)
            db_write = Database(path / named)
            root = path

        def exists_not_root(roc):
            roc = Path(roc)
            if not roc.exists() or not (roc / named).exists():
                return False
            return root.resolve() != roc.resolve() if root is not None else True

        if config['fixed'].exists() and config['fixed'].get() is not None:
            more = [Path(c) for c in config['fixed'].as_str_seq() if exists_not_root(c)]
            for m in more:
                db_fixed.append(Database(m / named, readonly=True))

        if db_write is not None and db_write.readonly:
            raise ValueError("Specified writable database location is readonly")
        if db_write is None:
            from platformdirs import user_cache_path
            db_write = Database(user_cache_path('restage', 'ess', ensure_exists=True) / named)
        if root is None:
            from platformdirs import user_data_path
            root = user_data_path('restage', 'ess')
        return cls(root, tuple(db_fixed), db_write)

    def query(self, method, *args, **kwargs):
        q = [x for r in self.db_fixed for x in getattr(r, method)(*args, **kwargs)]
        q.extend(getattr(self.db_write, method)(*args, **kwargs))
        return q

    def insert(self, method, *args, **kwargs):
        getattr(self.db_write, method)(*args, **kwargs)

    def query_instr_file(self, *args, **kwargs):
        query = [x for r in self.db_fixed for x in r.query_instr_file(*args, **kwargs)]
        query.extend(self.db_write.query_instr_file(*args, **kwargs))
        return query

    def insert_instr_file(self, *args, **kwargs):
        self.db_write.insert_instr_file(*args, **kwargs)

    def query_simulation_table(self, *args, **kwargs):
        return self.query('query_simulation_table', *args, **kwargs)

    def retrieve_simulation_table(self, *args, **kwargs):
        return self.query('retrieve_simulation_table', *args, **kwargs)

    def insert_simulation_table(self, *args, **kwargs):
        self.insert('insert_simulation_table', *args, **kwargs)

    def insert_simulation(self, *args, **kwargs):
        # By definition, 'self.db_write' is writable and Database.insert_simulation
        # _always_ ensures the presence of the specified table in its database.
        # Therefore this method 'just works'.
        self.insert('insert_simulation', *args, **kwargs)

    def retrieve_simulation(self, table_id: str, row: SimulationEntry):
        matches = []
        for db in self.db_fixed:
            if len(db.retrieve_simulation_table(table_id, False)) == 1:
                matches.extend(db.retrieve_simulation(table_id, row))
        if len(self.db_write.retrieve_simulation_table(table_id, False)) == 1:
            matches.extend(self.db_write.retrieve_simulation(table_id, row))
        return matches



FILESYSTEM = FileSystem.from_config('database')


def module_data_path(sub: str):
    path = FILESYSTEM.root / sub
    if not path.exists():
        path.mkdir(parents=True)
    return path


def directory_under_module_data_path(sub: str, prefix=None, suffix=None, name=None):
    """Create a new directory under the module's given data path, and return its path"""
    # Use mkdtemp to have a short-unique name if no name is given
    from tempfile import mkdtemp
    from pathlib import Path
    under = module_data_path(sub)
    if name is not None:
        p = under.joinpath(name)
        if not p.exists():
            p.mkdir(parents=True)
    return Path(mkdtemp(dir=under, prefix=prefix or '', suffix=suffix or ''))


def _compile_instr(entry: InstrEntry, instr: Instr, config: dict | None = None,
                   mpi: bool = False, acc: bool = False,
                   target=None, flavor=None):
    from mccode_antlr import __version__, Flavor
    from mccode_antlr.compiler.c import compile_instrument, CBinaryTarget
    if config is None:
        config = dict(default_main=True, enable_trace=False, portable=False, include_runtime=True,
                      embed_instrument_file=False, verbose=False)
    if target is None:
        target = CBinaryTarget(mpi=mpi or False, acc=acc or False, count=1, nexus=False)
    if flavor is None:
        flavor = Flavor.MCSTAS

    output = directory_under_module_data_path('bin')
    source_file = output.joinpath(instr.name).with_suffix('.c')
    binary_path = compile_instrument(instr, target, output, flavor=flavor, config=config, source_file=source_file)
    entry.mccode_version = __version__
    entry.binary_path = str(binary_path)
    return entry


def cache_instr(instr: Instr, mpi: bool = False, acc: bool = False, mccode_version=None, binary_path=None, **kwargs) -> InstrEntry:
    instr_contents = str(instr)
    # the query returns a list[InstrTableEntry]
    query = FILESYSTEM.query_instr_file(search={'file_contents': instr_contents, 'mpi': mpi, 'acc': acc})
    if len(query) > 1:
        raise RuntimeError(f"Multiple entries for {instr_contents} in {FILESYSTEM}")
    elif len(query) == 1:
        return query[0]

    instr_file_entry = InstrEntry(file_contents=instr_contents, mpi=mpi, acc=acc, binary_path=binary_path or '',
                                  mccode_version=mccode_version or 'NONE')
    if binary_path is None:
        instr_file_entry = _compile_instr(instr_file_entry, instr, mpi=mpi, acc=acc, **kwargs)

    FILESYSTEM.insert_instr_file(instr_file_entry)
    return instr_file_entry


def cache_get_instr(instr: Instr, mpi: bool = False, acc: bool = False) -> InstrEntry | None:
    query = FILESYSTEM.query_instr_file(search={'file_contents': str(instr), 'mpi': mpi, 'acc': acc})
    if len(query) > 1:
        raise RuntimeError(f"Multiple entries for {instr} in {FILESYSTEM}")
    elif len(query) == 1:
        return query[0]
    return None


def verify_table_parameters(table, parameters: dict):
    names = list(parameters.keys())
    if any(x not in names for x in table.parameters):
        raise RuntimeError(f"Missing parameter names {names} from {table.parameters}")
    if any(x not in table.parameters for x in names):
        raise RuntimeError(f"Extra parameter names {names} not in {table.parameters}")
    return table


def cache_simulation_table(entry: InstrEntry, row: SimulationEntry) -> SimulationTableEntry:
    query = FILESYSTEM.retrieve_simulation_table(entry.id)
    if len(query):
        for q in query:
            verify_table_parameters(q, row.parameter_values)
        table = query[0]
    else:
        table = SimulationTableEntry(list(row.parameter_values.keys()), f'pst_{entry.id}', entry.id)
        FILESYSTEM.insert_simulation_table(table)
    return table


def cache_has_simulation(entry: InstrEntry, row: SimulationEntry) -> bool:
    table = cache_simulation_table(entry, row)
    query = FILESYSTEM.retrieve_simulation(table.id, row)
    return len(query) > 0


def cache_get_simulation(entry: InstrEntry, row: SimulationEntry) -> list[SimulationEntry]:
    table = cache_simulation_table(entry, row)
    query = FILESYSTEM.retrieve_simulation(table.id, row)
    if len(query) == 0:
        raise RuntimeError(f"Expected 1 or more entry for {table.id} in {FILESYSTEM}, got none")
    return query


def cache_simulation(entry: InstrEntry, simulation: SimulationEntry):
    table = cache_simulation_table(entry, simulation)
    FILESYSTEM.insert_simulation(table, simulation)
